-- =============================================================================================================
-- *
-- * Copyright (c) Mike
-- *
-- * File Name: ram_256x16_sim.vhd
-- *
-- * Version: V1.0
-- *
-- * Release Date:
-- *
-- * Author(s): M.Freeman + others
-- *
-- * Description: simulation memory model
-- *
-- * Conditions of Use: THIS CODE IS COPYRIGHT AND IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND, INCLUDING,
-- *                    BUT NOT LIMITED TO, ANY IMPLIED WARRANTY OF MERCHANTABILITY AND FITNESS FOR A
-- *                    PARTICULAR PURPOSE.
-- *
-- * Notes: wrote this a while back, a blending of other designs, sorry lost original sources
-- *
-- =============================================================================================================

use std.textio.all;

Library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_arith.all;

entity ram_256x16_sim is
  generic ( 
    load_file_name     : string := "ram.dat";
    load_file_name_mem : string := "ram.mem";
    load_file_name_coe : string := "ram.coe";
    dump_file_name_coe : string := "dump.coe" );
  port (
    CLK      : in  std_logic;   
    ADDR_IN  : in  std_logic_vector( 7 downto 0 );
    DATA_IN  : in  std_logic_vector( 15 downto 0 );
    DATA_OUT : out std_logic_vector( 15 downto 0 );
    EN       : in  std_logic;     
    WE       : in  std_logic;
    DUMP     : in  std_logic
  );
end ram_256x16_sim;

architecture ram_256x16_sim_arch of ram_256x16_sim is
begin

  mem: process
    subtype word is std_logic_vector( 15 downto 0 );
    type mem_array is array (natural range 255 downto 0) of word;

    variable mem: mem_array;
    variable address : natural;
    variable L : line;
	 
	 variable saved : boolean := False;
	 
     --
     -- CONVERT NIBBLE TO CHAR
     --
    function hex_to_char (nibble: std_logic_vector(3 downto 0)) return character is
      variable hex: character;
    begin
      case nibble is
        when "0000" => hex := '0';
        when "0001" => hex := '1';
        when "0010" => hex := '2';
        when "0011" => hex := '3';
        when "0100" => hex := '4';
        when "0101" => hex := '5';
        when "0110" => hex := '6';
        when "0111" => hex := '7';
        when "1000" => hex := '8';
        when "1001" => hex := '9';
        when "1010" => hex := 'A';
        when "1011" => hex := 'B';
        when "1100" => hex := 'C';
        when "1101" => hex := 'D';
        when "1110" => hex := 'E';
        when "1111" => hex := 'F';
        when others => hex := 'X';
      end case;
      return hex;
    end hex_to_char;

    --
    -- LOAD MEM
    --
    procedure load( mem: inout mem_array) is
	 
      file load_data_file : text open READ_MODE  is load_file_name;
      file mem_data_file  : text open WRITE_MODE is load_file_name_mem;
      file coe_data_file  : text open WRITE_MODE is load_file_name_coe;		
		
      variable data_string : line;
      variable mem_string : line;
      variable coe_string : line;		
		
      variable addr, data, count : natural;
      variable data_char : character;
      variable valid : boolean;
		
      variable bin_addr : std_logic_vector(7 downto 0);
      variable bin_data : std_logic_vector(15 downto 0);		

    begin
      write( output, "ram initialization" );

      for addr in 0 to 255 
      loop
        mem( addr ) := "0000000000000000";
      end loop; 

      while not endfile( load_data_file ) 
      loop
        readline( load_data_file, data_string );
        read( data_string, addr, valid);        -- address	
		  
        if (valid)
        then       	  
          read( data_string, data_char );   -- space
			 
          data := 0; 
          count := 0;

          while count < 16
          loop
            read( data_string, data_char );
            if data_char /= ' ' 
            then	
              count := count + 1;     -- count bits processed
              data := data + data;    -- shift string
              if data_char = '1' 
              then 
                data :=data + 1;      -- insert 1 if set
              end if;  
            end if;		
          end loop;	  

          write( data_string , addr);
          write( data_string, ' ' );
          write( data_string, data);
          writeline( output, data_string );
          mem( addr ) := conv_std_logic_vector( data, 16 );
			 
          bin_addr := (conv_std_logic_vector( (addr*3), 8));
	  write( mem_string, '@');	 
          write( mem_string , hex_to_char( bin_addr(7 downto 4) ) );
          write( mem_string , hex_to_char( bin_addr(3 downto 0) ) );

          bin_data := conv_std_logic_vector( data, 16);			 
          write( mem_string , ' ');
          write( mem_string , hex_to_char( bin_data(3 downto 0) ) );	
          write( mem_string , hex_to_char( bin_data(7 downto 4) ) );
          write( mem_string , hex_to_char( bin_data(11 downto 8) ) );
          write( mem_string , hex_to_char( bin_data(15 downto 12) ) );
	  writeline( mem_data_file, mem_string );		 
        end if;		
      end loop;
			
      write( coe_string , "memory_initialization_radix = 16;");
      writeline( coe_data_file, coe_string );		 
      write( coe_string , "memory_initialization_vector = ");
      writeline( coe_data_file, coe_string );
		
      for addr in 0 to 255 
      loop
        bin_data :=  mem( addr );			 
        write( coe_string , ' ');		 
        write( coe_string , hex_to_char( bin_data(15 downto 12) ) );
        write( coe_string , hex_to_char( bin_data(11 downto 8) ) );
        write( coe_string , hex_to_char( bin_data(7 downto 4) ) );
        write( coe_string , hex_to_char( bin_data(3 downto 0) ) );	
        if addr = 255
	then
	  write( coe_string , " ;" );
	else
          write( coe_string , " ," ); 
        end if;
        writeline( coe_data_file, coe_string );

      end loop; 		
    end load;
	 
    --
    -- SAVE MEM
    --
    procedure save( mem: inout mem_array) is
      file coe_data_file  : text open WRITE_MODE is dump_file_name_coe;		

      variable coe_string : line;		
      variable addr : natural;	
      variable bin_data : std_logic_vector(15 downto 0);	

    begin
      write( coe_string , "memory_initialization_radix = 16;");
      writeline( coe_data_file, coe_string );		 
      write( coe_string , "memory_initialization_vector = ");
      writeline( coe_data_file, coe_string );
		
      for addr in 0 to 255 
      loop
        bin_data :=  mem( addr );			 
        write( coe_string , ' ');		 
        write( coe_string , hex_to_char( bin_data(15 downto 12) ) );
        write( coe_string , hex_to_char( bin_data(11 downto 8) ) );
        write( coe_string , hex_to_char( bin_data(7 downto 4) ) );
        write( coe_string , hex_to_char( bin_data(3 downto 0) ) );	
	if addr = 255
	then
	  write( coe_string , " ;" );
	else
          write( coe_string , " ," ); 
        end if;
	writeline( coe_data_file, coe_string );

      end loop;	
    end save;		

  --
  -- MAIN LOOP
  --	 
  begin
    load( mem );
    loop
      if (EN = '1') 
      then  
        address := conv_integer( ADDR_IN );        
        if (WE = '1') 
        then
          mem( address ) := DATA_IN;
        end if;
	     DATA_OUT <= mem( address );
      end if;
		  
      if (DUMP = '1' and not saved)
      then
        save( mem );
        saved := True;
      end if;
      wait on clk;

    end loop;
  end process;

end ram_256x16_sim_arch;


